from tkinter import *
import tkinter.simpledialog
import configparser
import os

class SettingsDialog(tkinter.simpledialog.Dialog):

    def body(self, master):
        self.inifile=getConfig()
        settings=self.inifile["settings"]
        f=("Helvetica",12)
        Label(master, text="Server:",font=f).grid(row=0)
        Label(master, text="Connect:",font=f).grid(row=1)
        self.e1var = StringVar(value=settings.get("Server","example.server.net:8922"))
        self.e2var = StringVar(value=settings.get("Connect","login command here"))
        self.e1 = Entry(master,width=60,font=f,textvariable=self.e1var)
        self.e2 = Entry(master,width=60,font=f,textvariable=self.e2var)
        self.fontvar=StringVar()
        Label(master,text="Font Size:",font=f).grid(row=2)
        self.spinbox=Spinbox(master,font=f,values=("6","8","10","12","14","16","18","20"),textvariable=self.fontvar)
        self.fontvar.set(settings.get("fontsize","12"))
        self.e1.grid(row=0, column=1)
        self.e2.grid(row=1, column=1)
        self.spinbox.grid(row=2,column=1,sticky=W)

        return self.e1 # initial focus

    def apply(self):
        first = self.e1.get()
        second = self.e2.get()
        print(first, second)
        self.result=(first,second)
        settings=self.inifile["settings"]
        settings["Server"]=first
        settings["Connect"]=second
        settings["Fontsize"]=self.fontvar.get()
        saveConfig(self.inifile)

def getConfigFile():
    return os.path.expanduser("~/moocoderpy.ini")

def getConfig():
    inifile=configparser.RawConfigParser()
    inifile.read(getConfigFile())
    if not inifile.has_section("settings"):
        inifile.add_section("settings")
    if inifile.has_section("DEFAULT"):
        inifile.remove_section("DEFAULT")
    if not inifile.has_section("test"):
        inifile.add_section("test")
    if not inifile.has_section("connections"):
        inifile.add_section("connections")
    return inifile

def saveConfig(inifile):
    with open(getConfigFile(),"w") as fp:
        inifile.write(fp)

if __name__ == "__main__":
    
    root=Tk()
    d=SettingsDialog(None,title="MoocoderPy Settings")
    print(d.result)


from gettext import find
from tkinter import *
import tkinter.simpledialog

class FindReplaceSettings:
    search:str=""
    replace:str=""
    caseSensitive:bool=False
    backward:bool=False
    isreplace:bool=False
    selection:bool=False
    wordmatch:bool=False
    all:bool=False
    go:bool=False

findReplaceSettings=FindReplaceSettings() # Global holder for settings.

class FindReplaceDialog(tkinter.simpledialog.Dialog):
    searchvar:StringVar
    replacevar:StringVar
    casevar:IntVar
    backvar:IntVar
    wordmatch:IntVar
    selection:IntVar
    all:IntVar
    backbtn:Checkbutton
    selbtn:Checkbutton

    def body(self, master):
        f=("Arial",12)
        Label(master, text="Search",font=f).grid(row=0,column=0)
        self.searchvar=StringVar(value=findReplaceSettings.search)
        self.search=Entry(master,width=60,font=f,textvariable=self.searchvar)
        self.search.grid(row=0, column=1,columnspan=3)
        if  findReplaceSettings.isreplace:
            Label(master, text="Replace",font=f).grid(row=1,column=0)
            self.replacevar=StringVar(value=findReplaceSettings.replace)
            self.replace=Entry(master,width=60,font=f,textvariable=self.replacevar)
            self.replace.grid(row=1, column=1, columnspan=3)
        self.casevar=IntVar(value=int(findReplaceSettings.caseSensitive))
        Checkbutton(master,text="Case Senstive",variable=self.casevar,font=f).grid(row=2,column=0,sticky=W)
        self.backvar=IntVar(value=int(findReplaceSettings.backward))
        bck=Checkbutton(master,text="Backwards",variable=self.backvar,font=f)
        bck.grid(row=2,column=1,sticky=W)
        self.backbtn=bck
        self.wordmatch=IntVar(value=int(findReplaceSettings.wordmatch))
        Checkbutton(master,text="Word Match",variable=self.wordmatch,font=f).grid(row=2,column=2,sticky=W)
        if findReplaceSettings.isreplace:
            self.all=IntVar(value=int(findReplaceSettings.all))
            all=Checkbutton(master,text="All",variable=self.all,font=f,command=self.updateScreen)
            all.grid(row=3,column=0,sticky=W)
            self.selection=IntVar(value=int(findReplaceSettings.selection))
            self.selbtn=Checkbutton(master,text="Selection",variable=self.selection,font=f)
            self.selbtn.grid(row=3,column=1,sticky=W)
            self.updateScreen()
        findReplaceSettings.go=False
        self.search.select_range(0, END)
        return self.search # initial focus

    def updateScreen(self,event=None):
        self.backbtn.configure(state=DISABLED if self.all.get() else NORMAL)
        self.selbtn.configure(state=DISABLED if not self.all.get() else NORMAL)


    def apply(self):
        findReplaceSettings.search=self.searchvar.get()
        if findReplaceSettings.isreplace:
            findReplaceSettings.replace=self.replacevar.get()
            findReplaceSettings.selection=self.selection.get()
            findReplaceSettings.all=self.all.get()
        findReplaceSettings.caseSensitive=bool(self.casevar.get())
        findReplaceSettings.backward=bool(self.backvar.get())
        findReplaceSettings.wordmatch=bool(self.wordmatch.get())
        findReplaceSettings.go=True

if __name__=="__main__":
    root=Tk()
    findReplaceSettings.isreplace=True
    FindReplaceDialog(None,"Find and Replace")
    print("Replace done.")
    print(findReplaceSettings.replace)
    print(findReplaceSettings.search)
    print(findReplaceSettings.caseSensitive)
    print(findReplaceSettings.backward)
    print(findReplaceSettings.go)
    print(findReplaceSettings.isreplace)




"""Some general purpose useful routines"""

def parse(s:str):
    """Get the next space separated word from a string. Return (word,remainder)"""
    s=s.strip()
    i=s.find(" ")
    if i<0:
        i=len(s)
    result=s[0:i].strip()
    remain=s[i+1:].strip()
    return (result,remain)

def parsesep(s:str, sep:str=","):
    """Get the next word from a string, separated ny sep. Return (word,remainder)"""
    i=s.find(sep)
    if i<0:
        i=len(s)
    result=s[0:i]
    remain=s[i+len(sep):]
    return (result,remain)

def getsepfield(s:str, n:int, sep:str=","):
    """Get nth word in s, separated by sep"""
    list=s.split(sep)
    if n<0 or n>=len(list):
        return ""
    return list[n]

def getfield(s:str, n:int):
    """Get nth word in s, separated by whitespace."""
    list=s.split()
    if n<0 or n>=len(list):
        return ""
    return list[n]

def atol(s:str)->int:
    """Convert string to int in a relaxed fashion... ie, stop when non-integer found."""
    result=""
    for c in s:
        if c=="-" and result=="":
            result+=c
        elif c==" ":
            continue
        elif c.isdigit():
            result+=c
        else:
            break
    try:    
        result=int(result)
    except:
        result=0
    return result

def ansiSameText(s1:str,s2:str)->bool:
    """Compare two strings case insensitive"""
    return (s1.lower()==s2.lower())

if __name__=="__main__":
    print(atol("1234x"))
    print(atol("-234 this is a test"))


from tkinter import simpledialog
from tkinter.ttk import Notebook
from ScrollText import *
from tkinter import *
from tkinter import messagebox,font
import socket, threading,select
import SettingsDialog
from codetext import *
from wgplib import *
import re,os
import mydialogs
import splitlist


class TerminalWindow(ScrollText):
    pages:Notebook=None
    getstack:bool=False
    setstackvisible=None
    stack:Text=None
    errorverb:str=""
    arglist={}
    verblist=[]
    proplist={}
    namelist={}
    lastobj:str=""
    lvVerbs:ttk.Treeview=None
    lvProperties:ttk.Treeview=None
    dumpobject:str=""
    normalfont:font.Font=None
    testtab:str=""
    lastlno:int=0
    lastproperty:str=""
    sendEntry:Entry = None
    filterLines:bool=False
    doboth:bool=False
    imgList={}

    ColorTable = (
        "#000000",
        "#c00000",
        "#00c000",
        "#c0c000",
        "#0000c0",
        "#c000c0",
        "#00c0c0",
        "#c0c0c0",
    )
    ColorTableBold = (
        "#000000",
        "#ff0000",
        "#00ff00",
        "#ffff00",
        "#0000ff",
        "#ff00ff",
        "#00ffff",
        "#ffffff",
    )
    ColorX256 = (
        "#000000",
        "#800000",
        "#008000",
        "#808000",
        "#000080",
        "#800080",
        "#008080",
        "#c0c0c0",
        "#808080",
        "#ff0000",
        "#00ff00",
        "#ffff00",
        "#0000ff",
        "#ff00ff",
        "#00ffff",
        "#ffffff",
        "#000000",
        "#00005f",
        "#000087",
        "#0000af",
        "#0000d7",
        "#0000ff",
        "#005f00",
        "#005f5f",
        "#005f87",
        "#005faf",
        "#005fd7",
        "#005fff",
        "#008700",
        "#00875f",
        "#008787",
        "#0087af",
        "#0087d7",
        "#0087ff",
        "#00af00",
        "#00af5f",
        "#00af87",
        "#00afaf",
        "#00afd7",
        "#00afff",
        "#00d700",
        "#00d75f",
        "#00d787",
        "#00d7af",
        "#00d7d7",
        "#00d7ff",
        "#00ff00",
        "#00ff5f",
        "#00ff87",
        "#00ffaf",
        "#00ffd7",
        "#00ffff",
        "#5f0000",
        "#5f005f",
        "#5f0087",
        "#5f00af",
        "#5f00d7",
        "#5f00ff",
        "#5f5f00",
        "#5f5f5f",
        "#5f5f87",
        "#5f5faf",
        "#5f5fd7",
        "#5f5fff",
        "#5f8700",
        "#5f875f",
        "#5f8787",
        "#5f87af",
        "#5f87d7",
        "#5f87ff",
        "#5faf00",
        "#5faf5f",
        "#5faf87",
        "#5fafaf",
        "#5fafd7",
        "#5fafff",
        "#5fd700",
        "#5fd75f",
        "#5fd787",
        "#5fd7af",
        "#5fd7d7",
        "#5fd7ff",
        "#5fff00",
        "#5fff5f",
        "#5fff87",
        "#5fffaf",
        "#5fffd7",
        "#5fffff",
        "#870000",
        "#87005f",
        "#870087",
        "#8700af",
        "#8700d7",
        "#8700ff",
        "#875f00",
        "#875f5f",
        "#875f87",
        "#875faf",
        "#875fd7",
        "#875fff",
        "#878700",
        "#87875f",
        "#878787",
        "#8787af",
        "#8787d7",
        "#8787ff",
        "#87af00",
        "#87af5f",
        "#87af87",
        "#87afaf",
        "#87afd7",
        "#87afff",
        "#87d700",
        "#87d75f",
        "#87d787",
        "#87d7af",
        "#87d7d7",
        "#87d7ff",
        "#87ff00",
        "#87ff5f",
        "#87ff87",
        "#87ffaf",
        "#87ffd7",
        "#87ffff",
        "#af0000",
        "#af005f",
        "#af0087",
        "#af00af",
        "#af00d7",
        "#af00ff",
        "#af5f00",
        "#af5f5f",
        "#af5f87",
        "#af5faf",
        "#af5fd7",
        "#af5fff",
        "#af8700",
        "#af875f",
        "#af8787",
        "#af87af",
        "#af87d7",
        "#af87ff",
        "#afaf00",
        "#afaf5f",
        "#afaf87",
        "#afafaf",
        "#afafd7",
        "#afafff",
        "#afd700",
        "#afd75f",
        "#afd787",
        "#afd7af",
        "#afd7d7",
        "#afd7ff",
        "#afff00",
        "#afff5f",
        "#afff87",
        "#afffaf",
        "#afffd7",
        "#afffff",
        "#d70000",
        "#d7005f",
        "#d70087",
        "#d700af",
        "#d700d7",
        "#d700ff",
        "#d75f00",
        "#d75f5f",
        "#d75f87",
        "#d75faf",
        "#d75fd7",
        "#d75fff",
        "#d78700",
        "#d7875f",
        "#d78787",
        "#d787af",
        "#d787d7",
        "#d787ff",
        "#d7af00",
        "#d7af5f",
        "#d7af87",
        "#d7afaf",
        "#d7afd7",
        "#d7afff",
        "#d7d700",
        "#d7d75f",
        "#d7d787",
        "#d7d7af",
        "#d7d7d7",
        "#d7d7ff",
        "#d7ff00",
        "#d7ff5f",
        "#d7ff87",
        "#d7ffaf",
        "#d7ffd7",
        "#d7ffff",
        "#ff0000",
        "#ff005f",
        "#ff0087",
        "#ff00af",
        "#ff00d7",
        "#ff00ff",
        "#ff5f00",
        "#ff5f5f",
        "#ff5f87",
        "#ff5faf",
        "#ff5fd7",
        "#ff5fff",
        "#ff8700",
        "#ff875f",
        "#ff8787",
        "#ff87af",
        "#ff87d7",
        "#ff87ff",
        "#ffaf00",
        "#ffaf5f",
        "#ffaf87",
        "#ffafaf",
        "#ffafd7",
        "#ffafff",
        "#ffd700",
        "#ffd75f",
        "#ffd787",
        "#ffd7af",
        "#ffd7d7",
        "#ffd7ff",
        "#ffff00",
        "#ffff5f",
        "#ffff87",
        "#ffffaf",
        "#ffffd7",
        "#ffffff",
        "#080808",
        "#121212",
        "#1c1c1c",
        "#262626",
        "#303030",
        "#3a3a3a",
        "#444444",
        "#4e4e4e",
        "#585858",
        "#626262",
        "#6c6c6c",
        "#767676",
        "#808080",
        "#8a8a8a",
        "#949494",
        "#9e9e9e",
        "#a8a8a8",
        "#b2b2b2",
        "#bcbcbc",
        "#c6c6c6",
        "#d0d0d0",
        "#dadada",
        "#e4e4e4",
        "#eeeeee",
    )

    def history(self,event):
        if event.keysym=="Up":
            self.historyidx+=1
            if self.historyidx>=len(self.historylst):
                self.historyidx=0
        elif event.keysym=="Down":
            self.historyidx-=1
            if self.historyidx<0:
                self.historyidx=len(self.historylst)-1
        if (self.historyidx>=0 and self.historyidx<len(self.historylst)):
            self.mytext.set(self.historylst[self.historyidx])

    def __init__(self, parent, **kwargs):
        super().__init__(parent, **kwargs)
        self.bottom = Frame(self, bg="LightGray")
        self.bottom.pack(side=BOTTOM, fill=X)
        self.sendbtn = Button(self.bottom, text="Snd", command=self.doSend, font=self.normalfont)
        self.mytext = StringVar()
        self.sendEntry = Entry(self.bottom, textvariable=self.mytext, font=self.fontctl)
        self.sendbtn.pack(side=RIGHT)
        self.sendEntry.pack(fill=X, expand=True)
        self.sendEntry.bind("<Return>",self.doSend)
        self.sendEntry.bind("<Up>", self.history)
        self.sendEntry.bind("<Down>", self.history)
        self.textbox.bind("<Double-Button-1>",self.dblClick)
        self.textbox.configure(exportselection=False,inactiveselectbackground=self.textbox["selectbackground"])
        self.textbox.bind("<KeyPress>",self.doRefocus)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.iscsi = False
        self.isesc = False
        self.parameters = []
        self.currentparam = ""
        self.ResetStyle()
        self.output=""
        self.currenttag=""
        self.taglist = []
        self.settag()
        self.listenthread=None
        self.connectString=""
        self.stopping=False
        self.lastline=""
        self.capturestr=""
        self.capturemode=0
        self.external_edit=""
        self.verbcollect=''
        self.onExamineLine=None
        self.upload=""
        self.historylst=[]
        self.historyidx=0
        self.loadIcons()
        ifile=SettingsDialog.getConfig()
        if ifile.has_section("history"):
            for (key,value) in ifile.items("history"):
                if key.isnumeric():
                    self.historylst.append(value)
            self.historyidx=-1

    def setFontSize(self,newsize:int):
        super().setFontSize(newsize)
        if self.sendEntry:
            self.sendEntry.configure(font=self.fontctl)


    def dblClick(self,event:Event):
        """Respond to double click"""
        self.gotoError(self.textbox)

    def settag(self):
        tagname = self.myfontcolor + ";" + self.mycolor
        if (tagname!=self.currenttag):
            self.flush()
        if not (tagname in self.taglist):
            cfg = {"background": self.mycolor, "foreground": self.myfontcolor}
            self.textbox.tag_config(tagname, cfg)
            self.taglist.append(tagname)
            self.textbox.tag_raise(SEL) # Make sure selection overrides.
        self.currenttag = tagname
    
    def doRefocus(self,event:Event):
        if len(event.char)>0 and event.char.isprintable():
            self.sendEntry.focus()
            self.mytext.set(self.mytext.get()+event.char)
            self.sendEntry.icursor(END)
            return "break"
        elif event.keysym=="Return":
            self.doSend(event)
            return "break"

    def parseVerb(self,value):
        """Return (obj,verb), or False if not valid."""
        if value.startswith('@'):
             value=parse(value)[1]
        (obj,value)=parsesep(value,':')
        (verb,value)=parse(value)
        if verb.startswith('"'):
             verb=getsepfield(verb,1,'"')
        i=verb.find("*")
        if (i>=0):
            verb=verb[0:i]
        if (verb!='') and (obj!=''):
            return (obj,verb)
        return False

    def addtext(self, msg):
        # print("Reading ",len(msg))
        for c in msg:
            self.addchar(c)
        self.textbox.see("end")

    def rgb(self, r, g, b):
        return "#%02x%02x%02x" % (r, g, b)
    
    def flush(self):
        if self.output!="":
            self.textbox.insert("end",self.output,self.currenttag)
            self.output=""
        #    self.textbox.see("end")

    def saveSettings(self):
        ifile=SettingsDialog.getConfig()
        if (ifile.has_section("history")):
            ifile.remove_section("history")
        ifile.add_section("history")
        h=ifile["history"]
        for i in range(10):
            if i>=len(self.historylst):
                break
            h[str(i)]=self.historylst[i]
        ifile["settings"]['LastDump']=self.dumpobject
        SettingsDialog.saveConfig(ifile)

    def ResetStyle(self):
        self.isbold = False
        self.isunderline = False
        self.isitalic = False
        self.isstrikeout = False
        self.myfontcolor = "#c0c0c0"
        self.mycolor = "#000000"

    def setbold(self, bold):
        if bold != self.isbold:
            if bold:
                ix = self.ColorTable.index(self.myfontcolor)
                if ix >= 0:
                    self.myfontcolor = self.ColorTableBold[ix]
                else:
                    self.myfontcolor = "#ffffff"
            else:
                ix = self.ColorTableBold.index(self.myfontcolor)
                if ix >= 0:
                    self.myfontcolor = self.ColorTable[ix]
                else:
                    self.myfontcolor = "#c0c0c0"
        self.isbold = bold

    def setattributes(self):
        extended = 0
        secondary = 0
        back = False
        xrgb = [0, 0, 0]
        for n in self.parameters:
            if extended > 0:
                if extended == 1:
                    secondary = n
                elif secondary == 5:  # X256
                    if back:
                        self.mycolor = self.ColorX256[n & 0xFF]
                    else:
                        self.myfontcolor = self.ColorX256[n & 0xFF]
                    extended = 0
                    continue
                elif secondary == 2:
                    xrgb[extended - 2] = n & 0xFF
                    if extended == 4:
                        if back:
                            self.mycolor = self.rgb(xrgb[0], xrgb[1], xrgb[2])
                        else:
                            self.myfontcolor = self.rgb(xrgb[0], xrgb[1], xrgb[2])
                        extended = 0
                        continue
                extended += 1
                continue
            if n == 0:
                self.ResetStyle()
            elif n == 1:
                self.setbold(True)
            elif n == 2 or n == 22:
                self.setbold(False)
            elif n == 3:
                self.isitalic = True
            elif n == 4:
                self.isunderline = True
            elif n == 9:
                self.isstrikeout = True
            elif n >= 30 and n <= 37:
                if self.isbold:
                    self.myfontcolor = self.ColorTableBold[n % 10]
                else:
                    self.myfontcolor = self.ColorTable[n % 10]
            elif n == 38:
                back = False
                extended = 1
            elif n >= 40 and n <= 47:
                if self.isbold:
                    self.mycolor = self.ColorTableBold[n % 10]
                else:
                    self.mycolor = self.ColorTable[n % 10]
            elif n == 48:
                back = True
                extended = 1
        # print("fg=",self.myfontcolor," bg=",self.mycolor)
        self.settag()

    def addchar(self, c: str):
        if self.isesc:
            if self.iscsi:
                if c.isalpha():
                    self.iscsi = False
                    self.isesc = False
                    self.parameters.append(int(self.currentparam))
                    self.currentparam = ""
                    if c == "m":
                        self.setattributes()
                elif c.isnumeric():
                    self.currentparam += c
                elif c == ";":
                    self.parameters.append(int(self.currentparam))
                    self.currentparam = ""
            elif c == "[":
                self.iscsi = True
                self.parameters = []
                self.currentparam = ""
        elif c == "\x1b":
            self.isesc = True
        else:
            self.isesc = False
            self.iscsi = False
            self.currentparam = ""
            if (self.capturemode!=1 and not self.filterLines) and c!="\r":
                self.output+=c
        if (c=="\n"):
            self.processLine()
            if self.filterLines and not self.lastline.startswith("Seconds"):
                self.addln(self.lastline)
            self.lastline=""
            self.checkLength()
        elif c!="\r":
            self.lastline+=c

    def checkLength(self):
        """Stop window growing forever"""
        start=self.textbox.index("end -10000lines")
        if (start!="1.0"):
            self.textbox.delete("1.0","end -9000lines")
            # print("Trimming ",start)

    def processLine(self):
        if self.capturemode==0:
            if self.lastline.startswith("#$# edit"):
                self.external_edit=self.lastline
                self.capturemode=1
                self.capturestr=""
            elif self.onExamineLine:
                self.onExamineLine(self.lastline)
        elif self.capturemode==1:
            if self.capturestr!="":
                self.capturestr+="\n"
            self.capturestr+=self.lastline
            if (self.lastline=="."):
                self.capturemode=0
                (name,upload)=self.parseExternal()
                self.openEdit(name,upload,self.capturestr)

    def selectError(self,obj:str,verb:str,lno:int)->bool:
        tabs=self.pages.tabs()
        for i in range(2,len(tabs)):
            w=self.pages.nametowidget(tabs[i])
            if (isinstance(w,ScrollText)):
                re:Text=w.textbox
                lines=re.get("1.0","end").splitlines()
                if (len(lines)>0 and (lines[0]+" ").lower().find(obj+':'+verb+' ')>=0):
                    self.pages.select(i)
                    start=str(lno+1)+'.0'
                    end=str(lno+1)+".end"
                    re.see(start)
                    re.focus_set()
                    re.tag_add(SEL,start,end)
                    #pagesChange(self);
                    return True
        return False
    
    def findVerbHelp(self,obj:str,verb:str)->str:
        """Extract help line from verb"""
        re=self.findVerbEditor(obj,verb)
        result='No Help'
        if not(re):
            return result
        lines=re.lines()
        if len(lines)>=2:
            s=lines[1]
            if (s.startswith('"')):
                s=s[1:]
                s=s[0:len(s)-2]
                result=s
        return result

    def findVerbEditor(self,obj:str, verb:str):
        searchverb=getsepfield(verb,0,'*') # Handle wildcards.
        for i in self.pages.tabs():
            w=self.pages.nametowidget(i)
            if isinstance(w,ScrollText):
                re:ScrollText=w
                if (re.tabtype!=1):
                    continue
                s=re.textbox.get("1.0","1.end")
                ret=self.parseVerb(s)
                if (ret):
                    (aobj,averb)=ret
                    if ansiSameText(aobj,obj) and ansiSameText(averb,searchverb):
                        return re
        return None
    
    def findPage(self,title:str):
        """Return widget matching title, or None if not found."""
        for i in self.pages.tabs():
            tabname=self.pages.tab(i,"text")
            if tabname==title:
                return self.pages.nametowidget(i)
        return None

    def currentEditor(self)->Text:
        w=self.currentPage()
        if isinstance(w,ScrollText):
            return w.textbox
        return None

    def currentPage(self)->ScrollText:
        w=self.nametowidget(self.pages.select())
        if isinstance(w,ScrollText):
            return w
        return None
    
    def currentTest(self)->Entry:
        w=self.currentPage()
        if isinstance(w,CodeText):
            return w.test
        return None

    def addtarget(self,dest:Text, line:str):
        dest.insert("end",line+"\n")
        dest.see("end")

    def doCheckTest(self,line:str):
        """Check for stack trace messages"""
        if self.getstack: 
            self.addtarget(self.stack,line)
        # #540:test (this == #540), line 5:  Type mismatch (expected integer; got float)
        # #151:+attacks, line 9:  Verb not found: #548:energy_cast()
        # #151:+deploy deploy, line 28:  Range error
        if line.startswith('#') and line.find(', line')>=0:
            startline=line
            (prog,line)=parsesep(line,',')
            line=parse(line)[1] # Skip "line"
            (lno,line)=parsesep(line,':')
            if line=="" or not lno.isnumeric(): # Can be spoofed by @dump output.
                return
            (obj,verb)=self.parseVerb(prog)    #Should strip out trailing defs.
            error=line
            self.getstack=True
            self.stack.delete("1.0","end")
            self.addtarget(self.stack,'Stack')
            self.addtarget(self.stack,error)
            self.addtarget(self.stack,obj+':'+verb+', line '+lno)
            self.setstackvisible(True)
            self.errorverb=''
            if not self.selectError(obj,verb,int(lno)):
                self.errorverb=verb
                self.errorobj=obj
                self.lastlno=int(lno)
        elif (line=='(End of traceback)'):
            self.getstack=False
            if self.errorverb!='':
                self.findVerb(self.errorobj,self.errorverb,self.lastlno)
            self.errorverb=''
    
    def doCheckCompile(self,line:str)->None:
        """Check that verb has compiled"""
#        var lno,x,n:Integer; s1,s2:String;  e:TEdit;
#        adddebug(line);
        if (line.startswith('Line ')):
            (s1,line)=parsesep(line,':');
            lno=atol(s1[4:])
            self.stack.delete("1.0","end")
            self.stack.insert("1.0","Compile Error\n"+line)
            self.setstackvisible(True)
            re=self.currentEditor()
            ix=str(lno+1)+".0"
            re.mark_set(INSERT,ix)
            re.see(ix)
            re.tag_add(SEL,ix,str(lno+1)+".end")
            re.focus_set()
        elif line=='Verb not programmed.':
            self.onExamineLine=self.doCheckTest
        elif line=='Verb programmed.':
            self.onExamineLine=self.doCheckTest
            if not self.checkSendTest():
                self.showmessage(line)
    
    def checkSendTest(self)->bool:
        e=self.currentTest()
        if e and (e.get()!=""):
            self.sendCmd(e.get())
            self.testtab=self.pages.select()
            self.getstack=False
            ifile=SettingsDialog.getConfig()
            ifile["test"][self.currentPage().testName()]=e.get()
            SettingsDialog.saveConfig(ifile)
            self.onExamineLine=self.doCheckTest
            self.pages.select(self)
            return True
        else:
            return False

    def addTab(self,caption:str,text:str,tabtype:int):
        t=CodeText(self.pages,tabtype,background="black",foreground="white",font=("Courier",self.fontsize,"bold"),
                   insertbackground="white",nowrap=True)
        if tabtype==CodeText.MODE_CODE:
            img=self.imgList["verb"]
        elif tabtype==CodeText.MODE_PROPERTY:
            img=self.imgList["property"]
        else:
            if caption=="Scratchpad":
                img=self.imgList["scratchpad"]
            elif caption=="Local Edit":
                img=self.imgList["file"]
            else:
                img=self.imgList["edit"]
        self.pages.add(t,text=caption,compound=TOP,image=img)
        t.tw=self
        t.setText(text)
        t.caption=caption
        t.tabtype=tabtype
        ifile=SettingsDialog.getConfig()
        t.testvar.set(ifile["test"].get(t.testName(),""))
        return t
    
    def doRefresh(self,verbstr:str)->None:
        """Reload a verb"""
        try:
            (verb,obj)=self.parseVerb(verbstr)
            self.fetchVerb(verb,obj)
        except:
            pass

    def doRefreshProperty(self,propstr:str)->None:
        """Reload a property"""
        try:
            (obj,prop)=propstr.split(".")
        except:
            return
        self.sendCmd(";"+propstr)
        self.lastobj=obj
        self.lastproperty=propstr
        self.onExamineLine=self.doCheckProperty

    def doCheckProperty(self,line:str):
        """Examine line for property detail."""
        if not line.startswith('=> '):
            self.showmessage(line)
        else:
            line=line[3:]
            self.proplist[self.lastproperty]=line
            self.updateProperties(False)
            self.editProp(self.lastproperty)
            self.checkName(self.lastobj)
        self.onExamineLine=self.doCheckTest

    def doCheckVerb(self, line:str):
        if (line=='***finished***'):
            self.onExamineLine=self.doCheckTest
            t=self.verbcollect.splitlines()
            t.append(".")
            prog=t[0]
            if (prog=='That object does not define that verb.'):
                messagebox.showwarning("MooCoderPy",'Verb not found.')
                return
            for i in range(len(t)):
                if t[i].endswith('[normal]'): # Stupid ansi is stupid.
                    s=t[i]
                    t[i]=s[0:len(s)-len('[normal]')]
            progline=prog
            (obj,prog)=parsesep(prog,':')
            args=prog
            if (args.startswith('"')):
                args=args.replace('"','').strip()
            (verb,prog)=parse(prog)
            if (verb.startswith('"')):
                verb=getsepfield(verb,1,'"')
            i=verb.find('*')
            if (i>=0):
                verb=verb[0:i]
            t[0]='@program '+obj+':'+verb
            if self.selectError(obj,verb,self.lastlno):
                re=self.currentEditor()
                lastix=re.index(INSERT)
                re.delete("1.0","end")
                re.insert("1.0","\n".join(t))
                if (self.lastlno>0):
                    re.see(str(self.lastlno+1)+".0")
                else:
                    re.mark_set(INSERT,lastix)
                    re.see(lastix)
                try:
                    self.currentPage().highlight(True)
                except:
                    pass
            else:
                self.addTab(obj+':'+verb,"\n".join(t),CodeText.MODE_CODE)
                self.selectError(obj,verb,self.lastlno)
            self.lastlno=0
            self.checkName(obj)
            tag=obj+':'+verb
            self.arglist[tag]=args
            if not(tag in self.verblist):
                self.verblist.append(tag)
            try:                
                pg=self.currentPage().setLabel(progline)
            except:
                pass
            self.updateVerbs()
        else:
            self.verbcollect+=line+"\n"
    
    def findLocalEdit(self,name:str)->CodeText:
        """Find the local editor by name"""
        for tab in self.pages.tabs():
            w=self.pages.nametowidget(tab)
            if (isinstance(w,CodeText) and w.mode==CodeText.MODE_EDIT):
                if (w.caption==name):
                    return w
        return None

    def openEdit(self,name:str, upload:str, text:str)->CodeText:
        """Find and open a tab for a local edit"""
        re=self.findLocalEdit(name)
        if re:
            if re.upload!="##ScratchPad##":
                re.setText(text)
        else:
            re=self.addTab(name,text,CodeText.MODE_EDIT)
        re.upload=upload
        re.syntax=(upload.find("@program")>=0 or upload=="##ScratchPad" or name=="Local Edit")
        re.setLabel(upload)
        re.highlight()
        self.pages.select(re)
        return re
        
    def checkName(self,obj:str)->None:
        """Check that we know the name of an object."""
        if not(obj in self.namelist):
            self.sendCmd(';'+obj+'.name')
            self.onExamineLine=self.doCheckName
        self.lastobj=obj
    
    def doCheckName(self,line:str):
        """Examine stream for name of object."""
        if line.lower().find('***finished***')>=0:
            return # Ignore trailing stuff.
        if line.startswith('=> 0'):
            return # tailing result.
        self.onExamineLine=self.doCheckTest
        if not(line.startswith('=>')):
            self.addln('Name not found.')
            return
        aname=getsepfield(line,1,'"')
        self.namelist[self.lastobj]=aname
        self.updateVerbs()
        self.updateProperties(False)
    
    def updateVerbs(self)->None:
        """Update Verb list window"""
#        var nd:TListItem; s:String; obj,verb:String;
#            oldverb,oldobj:String; i:Integer;
#        lvVerbs.Items.BeginUpdate;
        oldverb=''
        nd=self.lvVerbs.item(self.lvVerbs.focus())
        if nd["text"]!="":
            oldobj=nd["text"].lower()
            oldverb=nd["values"][1].lower()
        for i in self.lvVerbs.get_children(): #Clear list
            self.lvVerbs.delete(i)
        self.verblist.sort() # Todo... make cleverer.
        for s in self.verblist:
            obj=getsepfield(s,0,':')
            verb=getsepfield(s,1,':')
            name=self.namelist[obj] if obj in self.namelist else obj
            args=self.arglist[s] if s in self.arglist else ""
            values=(name,verb,args,self.findVerbHelp(obj,verb))
            self.lvVerbs.insert("","end",text=obj,values=values)
        self.fitListContents(self.lvVerbs)
        if oldverb!="":
            for i in self.lvVerbs.get_children():
                nd=self.lvVerbs.item(i)
                if nd["text"].lower()==oldobj and nd["values"][1].lower()==oldverb:
                    self.lvVerbs.selection_set(i)
                    self.lvVerbs.see(i)
                    break

    def updateProperties(self,focus=True):
        """Update Property List"""
        oldprop=""
        nd=self.lvProperties.item(self.lvProperties.focus())
        if nd["text"]!="":
            oldobj=nd["text"].lower()
            oldprop=nd["values"][1].lower()
        for item in self.lvProperties.get_children():
            self.lvProperties.delete(item)
        for k in sorted(self.proplist.keys()):
            v=self.proplist[k]
            (obj,prop)=k.split(".", 1)
            name=self.namelist[obj] if obj in self.namelist else obj
            self.lvProperties.insert("",END,text=obj, values=(name,prop,v))
        self.fitListContents(self.lvProperties)
        if (focus):
            self.pages.select(self.lvProperties.winfo_parent())
        if oldprop!="":
            for i in self.lvProperties.get_children():
                nd=self.lvProperties.item(i)
                if nd["text"].lower()==oldobj and nd["values"][1].lower()==oldprop:
                    self.lvProperties.selection_set(i)
                    self.lvProperties.see(i)
                    break

    def fitListContents(self,alist:ttk.Treeview):
        cols=alist.cget("columns")
        colnames=["#0"]+list(cols)
        widths=[0 for x in colnames]
        for i in range(len(colnames)):
            n=colnames[i]
            widths[i]=max(widths[i],self.normalfont.measure(alist.heading(n,"text")))
        for iid in alist.get_children():
            line=alist.item(iid)
            widths[0]=max(widths[0],self.normalfont.measure(line["text"]))
            v=line["values"]
            for i in range(1,min(len(colnames),len(v)+1)):
                widths[i]=max(widths[i],self.normalfont.measure(v[i-1]))
        spacing=self.normalfont.measure("  ")
        widths[0]+=spacing*2 # Guessing at present: allow space for image
        for i in range(len(colnames)):
            alist.column(colnames[i],width=widths[i]+spacing)

    def parseExternal(self):
        try:
            t=self.external_edit.split(" name: ")[1]
            tt=t.split(" upload: ")
            return (tt[0].strip(),tt[1].strip())
        except:
            return("","")
    
    def doupdate(self,caption:str, upload:str,text:str):
        if upload=="" or caption=="Local Edit":
            messagebox.showwarning("Send Update","Not in @edit mode")
            return False
        if upload!="##ScratchPad##":
            self.sendCmd(upload)
        self.sendCmd(text)
        self.after(500,self.sendtext,"Update complete\n")
        e=self.currentTest()
        if e and (e.get()!=""):
            test=e.get()
            self.testtab=self.pages.select()
            self.getstack=False
            self.onExamineLine=self.doCheckTest
            ifile=SettingsDialog.getConfig()
            ifile["test"][self.currentPage().testName()]=e.get()
            SettingsDialog.saveConfig(ifile)
            self.after(1000,self.sendCmd,test)
        self.pages.select(self)
        return True
    
    def docompile(self,page:ScrollText):
        if not(isinstance(page, CodeText)):
            return
        if page.mode==CodeText.MODE_EDIT:
            self.doupdate(page.caption,page.upload,page.textbox.get("1.0","end"))
        elif page.mode==CodeText.MODE_PROPERTY:
            text=page.getText()
            if len(text)>0 and not(text[0] in "{["): # Not formatted as a list. Treat as list of strings.
                text=text[:-1] if  text.endswith("\n") else text
                text=text.split("\n")
                text=[line.replace('"',r'\"') for line in text]
                value='{"'+'","'.join(text)+'"}'
            else:
                value=splitlist.joinList(page.getText())
            self.sendCmd(page.upload+"="+value)
            self.proplist[page.caption]=value
            self.updateProperties(False)
            self.onExamineLine=self.doCheckUpdateProperty
        else:
            self.sendCmd(page.textbox.get("1.0","end"))
            self.onExamineLine=self.doCheckCompile

    def doSend(self,event:Event=None):
        s = self.mytext.get()
        self.mytext.set("")
        if self.socket != None:
            self.socket.send((s + "\n").encode("utf-8"))
            self.historylst.insert(0,s)
            self.historyidx=-1
        self.sendEntry.focus()

    def doConnect(self, server, port):
        self.disconnect()
        self.sendtext("Connecting to %s %d\n" % (server,port))
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((server, port))
        self.stopping=False
        self.listenthread = threading.Thread(target=self.doListen)
        self.listenthread.start()
        if (self.connectString!=""):
            self.after(1000,self.sendCmd,self.connectString)
        print("Connected")
    
    def showmessage(self,msg):
        messagebox.showinfo("MooCoderPy",msg)

    def loadVerb(self,verbdef):
        v=self.parseVerb(verbdef)
        if not v:
            self.showmessage("Invalid Verb Syntax")
            return
        (obj,verb)=v
        self.findVerb(obj,verb,0)

    def findVerb(self,obj:str,verb:str, lno:int):
        self.lastlno=lno
        if (obj=='#-1'):
            return # Not a real verb.
        #if not(self.selectError(obj,verb,lno)):
        self.fetchVerb(obj,verb)
    
    def fetchVerb(self,obj,verb):
        self.sendCmd('@list '+obj+':'+verb+' without numbers')
        self.sendCmd(';player:tell("***finished***")')
        self.verbcollect=''
        self.onExamineLine=self.doCheckVerb

    def sendCmd(self,cmd):
        buf=(cmd+"\n").encode("utf-8")
        self.socket.sendall(buf)
        

    def doConnectStr(self):
        if (self.connectString!=""):
            self.sendCmd(self.connectString)

    def disconnect(self):
        print("Disconnecting...")
        try:
            self.stopping=True
            self.socket.close()
            if (self.listenthread!=None):
                threading.join(self.listenthread)
            self.listenthread=None
        except:
            pass

    def doListen(self):
        try:
            while True:
                (readlist,writelist,exceptlist)=select.select([self.socket],[],[self.socket],1) 
                if self.stopping:
                    print("Listen thread stopping.")
                    break
                if self.socket in exceptlist:
                    print("Exception found.")
                    break
                if self.socket in readlist:
                    b = self.socket.recv(4096)
                    if len(b)==0: # Socket probably closed.
                        self.sendtext("\nDisconnected\n")
                        break
                    self.sendtext(b.decode("utf-8"))
        except Exception as err:
            print("Connection error: {0}".format(err))
            try:
                self.sendtext("Connection error: {0}".format(err))
            except:
                pass
        print("Listen thread done.")
        self.listenthread=None

    def getVerbs(self,event:Event=None):
        s=simpledialog.askstring('Verbs','Load Verb list for object:', initialvalue=self.dumpobject)
        if (s):
            self.dumpobject=s
            self.sendCmd('@verbs '+self.dumpobject)
            self.onExamineLine=self.doCheckVerbs
            self.doboth=False


    def getProperties(self,event:Event=None):
        s=simpledialog.askstring('Properties','Load Property list for object:', initialvalue=self.dumpobject)
        if (s):
            self.dumpobject=s
            self.sendCmd('@dump '+self.dumpobject+" with noverbs")
            self.onExamineLine=self.doCheckProperties
    
    def getBoth(self,event:Event=None):
        s=simpledialog.askstring('Verbs and Properties','Load Verb and Property list for object:', initialvalue=self.dumpobject)
        if (s):
            self.dumpobject=s
            self.sendCmd('@verbs '+self.dumpobject)
            self.doboth=True
            self.onExamineLine=self.doCheckVerbs

    def doCheckProperties(self,line:str):
        # ;;#151.("rank_chart") = {"-", "F", "D", "C", "B", "A", "S", "SS", "SSS", "SSSS", "SSSSS"}
        if line.startswith(";;#"):
            (k,v)=parsesep(line," = ")
            k=re.sub(r'["();]',"",k)
            self.proplist[k]=v
            self.lastobj=getsepfield(k,0,".")
            p=self.findPage(k)
            if p:
                p.setText(splitlist.splitList2(v))
        elif "***finished***" in line:
            self.onExamineLine=self.doCheckTest
            self.updateProperties()
            self.checkName(self.lastobj)

    def doCheckVerbs(self,line:str)->None:
        """Fetch verbs for an object"""
        self.onExamineLine=self.doCheckTest
        if not(line.lower().startswith(';verb')):
            messagebox.showerror("MooCoderPy","Verb not found.\n"+line)
            return
        line=parsesep(line,'(')[1]
        (obj,line)=parsesep(line,')')
        line=parsesep(line,'{')[1]
        line=line.strip()
        if line.endswith('}'):
            line=line[0:len(line)-1]
        self.verblist=[x for x in self.verblist if not(x.startswith(obj+":"))]    
        t=line.split(",")
        for s1 in t:
            s=s1.strip()[1:]
            s=s[0:len(s)-1]
            verb=getfield(s,0)
            self.verblist.append(obj+':'+verb);
        self.checkName(obj)
        self.updateVerbs()
        self.pages.select(self.lvVerbs.winfo_parent())
        if self.doboth:
            self.doboth=False
            self.sendCmd('@dump '+self.dumpobject+" with noverbs")
            self.onExamineLine=self.doCheckProperties

    
    def clearProject(self):
        """Close all open tabs and clear verb list"""
        if messagebox.askyesno("MooCoderPy","Clear this project?"):
            tabs=list(self.pages.tabs())
            tabs.reverse()
            for tab in tabs:
                x=self.pages.nametowidget(tab)
                if (hasattr(x,"tabtype") and x.tabtype==1):
                    x.close()
            self.verblist.clear()
            self.namelist.clear()
            self.arglist.clear()
            self.proplist.clear()
            self.updateVerbs()
            self.updateProperties()
            self.pages.select(self)

    def loadIcons(self):
        self.imgfolder=os.path.dirname(os.path.abspath(__file__))+"/img"
        self.imgList={}
        imagemap = [("terminal", "monitor"), ("verblist", "code"), ("verb", "inspect-code"), ("proplist", "edit-property"),
                    ("property", "new-property"),
                    ("scratchpad", "spiral-bound-booklet"), ("edit", "edit"),
                    ("file", "document")]
        for (name, image) in imagemap:
            self.imgList[name]=PhotoImage(file=self.imgfolder+"/icons8-"+image+"-24.png")

    def gotoError(self,text:Text):
        line=text.get(INSERT+" linestart",INSERT+" lineend").strip()
        if line.startswith('... called from'):
            line=line[line.find("#"):]   # Strip off leading stuff.
            (prog,line)=parsesep(line,',')
            line=parse(line)[1]
            lno=atol(line)
            v=self.parseVerb(prog)
            if v:
                self.findVerb(v[0],v[1],lno)
        elif line.startswith('#') and (', line' in line):
            (prog,line)=parsesep(line,',')
            line=parse(line)[1] # Skip "line"
            (tmp,line)=parsesep(line,':')
            lno=atol(tmp)
            v=self.parseVerb(prog)     # Should strip out trailing defs.
            if v:
                (obj,verb)=v
                if not self.selectError(obj,verb,lno):
                    self.fetchVerb(obj,verb)
    
    def editProp(self,propname:str):
        """Edit a property"""
        try:
            (obj,prop)=propname.split(".")
        except:
            messagebox.showwarning("MooCoderPy","Invalid property format.")
            return
        if not(propname in self.proplist):
            messagebox.showwarning("MooCoderPy","Unknown Property.")
            return
        value=self.proplist[propname]
        if value.startswith("[") or value.startswith("{"):
            myedit=self.findPage(propname)
            if not(myedit):
                myedit=self.addTab(propname,"",CodeText.MODE_PROPERTY)
            myedit.syntax=True
            myedit.setText(splitlist.splitList2(value))
            myedit.upload=";;"+propname
            myedit.setLabel(myedit.upload)
            self.pages.select(myedit)
            return
        newvalue=mydialogs.askstring("Edit "+propname,"New Value",initialvalue=value,width=80)
        if (not newvalue):
            return
        self.proplist[propname]=newvalue
        self.sendCmd(";;"+propname+"="+newvalue)
        self.updateProperties()
        self.onExamineLine=self.doCheckUpdateProperty
    
    def doCheckUpdateProperty(self,line:str):
        """Check on the results of an edit prop"""
        if not line.startswith('=> 0'):
            self.showmessage(line)
        else:
            if self.checkSendTest():
                return
        self.onExamineLine=self.doCheckTest

if __name__=="__main__":
    root=Tk()
    tw=TerminalWindow(root)
    tw.pack(fill=BOTH, expand=True)
    tw.sendtext("Mary had a little lamb.")
    root.mainloop()




import tkinter as tk
from tkinter import ttk
from tkinter import font
from queue import Queue
from threading import Thread
import time

# Turns out there's a perfectly good scrolledtext component already, but this works as a useful wrapper anyway.

class ScrollText(tk.Text):
    tabtype=0
    fontsize:int=12
    fontctl:font.Font=None
    nowrap:bool=False
    hscroll:tk.Scrollbar = None

    def __init__(self, parent, **kwargs):
        tk.Frame.__init__(self,parent)
        if "nowrap" in kwargs:
            self.nowrap=kwargs["nowrap"]
            del kwargs["nowrap"]
        self.textbox=tk.Text(self,**kwargs)
        self.scrollbar=ttk.Scrollbar(self, orient='vertical', command=self.textbox.yview)
        self.scrollbar.pack(side=tk.RIGHT,fill=tk.Y)
        self.textbox.pack(fill=tk.BOTH,expand=True)
        self.textbox["yscrollcommand"]=self.scrollbar.set
        if self.nowrap:
            self.textbox.configure(wrap=tk.NONE)
            self.hscroll=ttk.Scrollbar(self, orient=tk.HORIZONTAL, command=self.textbox.xview)
            self.hscroll.pack(side=tk.BOTTOM, fill=tk.X)
            self.textbox["xscrollcommand"]=self.hscroll.set

        self.queue=Queue(20)
        self.bind("<<sendtext>>",self.handletext)
        self.setFontSize(self.fontsize)
    
    def setFontSize(self,newsize:int):
        self.fontctl=font.Font(family="Courier", size=newsize, weight="bold")
        self.textbox.configure(font=self.fontctl)
        self.fontsize=newsize

    def sendtext(self,msg):
        self.queue.put(msg)
        self.event_generate("<<sendtext>>",when='tail')
    
    def addln(self,line:str):
        self.textbox.insert("end","\n"+line)
        self.textbox.see("end")
    
    def clear(self):
        self.textbox.delete("1.0","end")

    def handletext(self,event):
        while not self.queue.empty():
            msg=self.queue.get()
            self.addtext(msg)
            self.queue.task_done()
        self.flush()
        self.textbox.see("end")
    
    def flush(self):
        pass

    def addtext(self,msg):
            self.textbox.insert("end",msg)
            self.textbox.see("end")
    
    def getText(self)->str:
        return self.textbox.get("1.0",tk.END)

    def testSend(self):
        time.sleep(2)
        self.sendtext("This should appear after 5 seconds\nAnd this is a second line.")
        print("Thread done.")
    
    def lines(self):
        """Return text as a list of strings"""
        return self.textbox.get("1.0","end").split("\n")

if __name__ == "__main__":
    root=tk.Tk()
    wtest=ScrollText(root,background="blue",foreground="white")        
    wtest.pack()
    wtest.textbox.insert("1.0","Hello")
    t=Thread(target=wtest.testSend)
    t.start()
    wtest.mainloop()


from cgitb import text
from tkinter import messagebox, simpledialog
from ScrollText import *
from tkinter import *
from time import *
from findreplacedialog import *
import re

class CodeText(ScrollText):
    bottom:Frame
    test:Entry
    caption:str
    testvar:StringVar
    labelvar:StringVar
    modifying=False
    lastfind:str=""
    popup:Menu
    tw=None #Terminal window.
    mode:int=0
    syntax:bool=True
    syntaxlist=set()
    upload:str=""
    linenos:Label=None

    KEYWORDLIST=('if','then','else','elseif','for',
           'while','endfor','endwhile','endif',
           'try','except','finally','endtry','break','continue','in','fork','endfork')
    
    COLORLIST=("white","lawn green","cyan","magenta","orange","yellow")
    OPENBRACKETS=("(","{","[")
    CLOSEBRACKETS=(")","}","]")
    MODE_CODE = 1 # Code editor, from verb list, syntax highlight.
    MODE_EDIT = 2 # Local Edit 
    MODE_PROPERTY = 3 # Property editor (from Proplist)

    def __init__(self, parent, mode:int, **kwargs):
        super().__init__(parent, **kwargs)
        self.textbox.pack_forget()
        if self.hscroll:
            self.hscroll.pack_forget()
        self.labelvar=StringVar()
        lbl = Label(self,textvariable=self.labelvar,font="Arial 12")
        lbl.pack(side=TOP,fill=X)
        self.linenosvar=StringVar()
        self.linenos=Label(self,textvariable=self.linenosvar, font=self.textbox["font"], background="Silver", foreground="Black",width=4,justify=RIGHT)
        self.linenos.pack(side=LEFT, anchor=NE,padx=0)
        self.textbox.pack(fill=BOTH, expand=True)
        self.bottom = Frame(self, bg="LightGray")
        self.bottom.pack(side=BOTTOM, fill=X)
        if self.hscroll:
            self.hscroll.pack(side=BOTTOM, fill=X)
        self.posvar=StringVar()
        self.posvar.set("000.000")
        poslbl=Label(self.bottom,textvariable=self.posvar)
        poslbl.pack(side=LEFT)
        lbl=Label(self.bottom,text="Test ",font="Arial 12")
        lbl.pack(side=LEFT)
        self.testvar=StringVar()
        self.test=Entry(self.bottom,width=60,font="Arial 12",textvariable=self.testvar)
        self.test.pack(side=LEFT,fill=X,expand=True)
        self.setcolors()
        self.textbox.bind("<ButtonRelease>",self.trackLocation)
        self.textbox.bind("<KeyRelease>",self.trackLocation)
        self.textbox.bind("<Enter>",self.trackLocation)
        self.textbox.bind("<Control-Key-G>",self.gotoLine)
        self.textbox.bind("<Control-Key-g>",self.gotoLine)
        self.textbox.bind("<<Modified>>",self.modified)
        self.textbox.bind("<Control-Key-F>",self.find)
        self.textbox.bind("<Control-Key-f>",self.find)
        self.textbox.bind("<Control-Key-R>",self.refresh)
        self.textbox.bind("<Control-Key-r>",self.refresh)
        self.textbox.bind("<F3>",self.findAgain)
        self.textbox.bind("<<Paste>>",self.cutPaste)
        self.textbox.bind("<<Cut>>",self.cutPaste)
        self.mode=mode
        self.buildPopup()
        self.textbox.bind("<Button-3>",self.doPopup)
        self.textbox.bind("<Button-2>",self.doPopup) # For Mac, in theory.
        self.textbox.configure(undo=True,exportselection=False,inactiveselectbackground=self.textbox["selectbackground"])
        self.textbox.bind("<Return>",self.handleIndent)
        self.textbox["yscrollcommand"]=self.handleScroll

    def handleScroll(self,first,last):
        self.scrollbar.set(first,last)
        self.updateLineNos()

    def handleIndent(self,event:Event):
        line=self.textbox.get("insert linestart","insert lineend")
        match = re.match(r'^(\s+)', line)
        current_indent = len(match.group(0)) if match else 0
        self.textbox.insert(INSERT,"\n"+(" " * current_indent))
        return "break"
    
    def setFontSize(self, newsize: int):
        super().setFontSize(newsize)
        if self.linenos:
            self.linenos.configure(font=self.fontctl)

    def cutPaste(self,event:Event):
        """Trigger highlight on cut and paste."""
        self.textbox.after(0,self.highlight,True)

    def buildPopup(self):
        """Make the Popup menu"""
        p=Menu(self,tearoff=0,font="Arial 12")
        p.add_command(label="Goto Line Ctrl+G",command=self.gotoLine,underline=0)
        p.add_command(label="Find Ctrl+F",command=self.find,underline=0)
        p.add_command(label="Find Again F3",command=self.findAgain,underline=6)
        p.add_command(label="Replace Ctrl+H",command=self.doReplace, underline=1)
        self.textbox.bind("<Control-H>",self.doReplace)
        self.textbox.bind("<Control-h>",self.doReplace)
        p.add_command(label="Refresh Ctrl+R",command=self.refresh,underline=0)
        p.add_command(label="Undo Last Edit Ctrl+Z",command=self.undo)
        self.textbox.bind("<Control-z>",self.undo)
        self.textbox.bind("<Control-Z>",self.undo)
        p.add_command(label="Redo Last Edit Shift+Ctrl+Z",command=self.redo)
        self.textbox.bind("<Shift-Control-z>",self.redo)
        self.textbox.bind("<Shift-Control-Z>",self.redo)
        p.add_command(label="Redo Syntax",command=lambda: self.highlight(True),underline=10)
        p.add_command(label="Close",command=self.close,underline=0)
        if self.mode==CodeText.MODE_PROPERTY:
            p.add_command(label="Edit as Text",command=self.flattenText,underline=3)
        self.popup=p
    
    def warn(self,text:str):
            messagebox.showwarning("MooCoderPy",text)

    def flattenText(self):
        text=self.getText().split("\n")
        if len(text)<2 or not text[0].startswith("{"):
            self.warn("Not a list.")
            return
        text=text[1:-1]
        if text[-1]=="}":
            text=text[:-1]
        for (k,line) in enumerate(text):
            line=line.strip()
            if line.startswith('"'):
                line=line[1:]
            if line.endswith(","):
                line=line[:-1]
            if line.endswith('"'):
                line=line[:-1]
            line=line.replace(r'\"','"')
            text[k]=line
        self.syntax=False
        self.setText("\n".join(text))

    def close(self):
        """Close this window"""
        self.destroy()

    def setLabel(self,atext):
        self.labelvar.set(atext)

    def undo(self,event:Event=None):
        try:
            self.textbox.edit_undo()
        except:
            pass

    def redo(self,event:Event=None):
        try:
            self.textbox.edit_redo()
        except:
            pass

    def setText(self,text):
        """Populate text. Does highlighting etc"""
        self.textbox.delete("1.0",END)
        self.textbox.insert("1.0",text)
        self.textbox.edit_reset()
        self.highlight(True)

    def doPopup(self,event:Event):
        """Do the actual popup"""
        try:
            self.popup.post(event.x_root,event.y_root)
        finally:
            self.popup.grab_release()
    
    def refresh(self,event=None):
        """Refresh from server"""
        if not self.tw:
            messagebox.showwarning("MooCoderPy","No Connection to Server")
            return
        if (self.mode==self.MODE_EDIT):
            if self.upload=="":
                messagebox.showerror("MooCoderPy","No upload command.")
            else:
                self.tw.sendCmd("@edit "+self.caption)
        elif self.mode==self.MODE_PROPERTY:
            self.tw.doRefreshProperty(self.caption)
        else:
            self.tw.doRefresh(self.caption)
        return "break"

    def currentLine(self)->int:
        """Return currently selected line no"""
        return int(self.textbox.index(INSERT).split(".")[0])

    def modified(self, event):
        if not(self.modifying): # Avoid recursion
            try:
                self.modifying=True
                lno=self.currentLine()
                self.syntaxHighlight(lno-1)
                self.syntaxHighlight(lno)
                self.syntaxHighlight(lno+1)
                # Set 'modified' to 0.  This will also trigger the <<Modified>>
                # virtual event which is why we need the sentinel.
                self.textbox.edit_modified(False)
            finally:
                # Clean the sentinel.
                self.modifying = False

    def trackLocation(self,event=None):
        """Update Cursor Location"""
        loc=self.textbox.index(INSERT)
        (x,y)=loc.split(".")
        offset=1 if self.tabtype==CodeText.MODE_CODE else 0
        s="%4d:%3d" % (int(x)-offset,int(y))
        self.posvar.set(s)
        if (self.lastfind!="" and self.lastfind!=loc):
            self.textbox.tag_remove("found","1.0",END)
        self.bracketMatching()

    def updateLineNos(self):
        topix=int(self.textbox.index("@0,0").split(".")[0])
        height=int(self.textbox.winfo_height()/self.fontctl.metrics("linespace"))
        offset=1 if self.tabtype==CodeText.MODE_CODE else 0
        self.linenosvar.set("\n".join([str(x+topix-offset) for x in range(height)]))
        self.highlight()

    def bracketMatching(self):
        """Match bracket types"""
        ix=self.textbox.index(INSERT)
        c=self.textbox.get(ix)
        self.textbox.tag_remove("bracket","1.0",END)
        if c in self.OPENBRACKETS:
            text=self.textbox.get(ix,END)
            bracketlen=self.bracketMatch(c,text,True)
            if bracketlen>0:
                self.textbox.tag_add("bracket",ix,("%s +%d chars" % (ix,bracketlen)))
        elif c in self.CLOSEBRACKETS:
            text=self.textbox.get("1.0",ix+" +1 chars")
            bracketlen=self.bracketMatch(c,text,False)
            if bracketlen>0:
                self.textbox.tag_add("bracket","%s -%d chars" % (ix,bracketlen-1),ix+" +1 chars")

    
    def bracketMatch(self,bracket:str,text:str, forwards:bool)->int:
        """Routine to actually track the brackets. Returns the number of characters, 0 if not found. """
        if forwards:
            close=self.CLOSEBRACKETS[self.OPENBRACKETS.index(bracket)]
        else:
            close=self.OPENBRACKETS[self.CLOSEBRACKETS.index(bracket)]
            text=text[::-1] # Reverse
        result=0
        match=0
        for xx in text:
            result+=1
            if (xx==bracket):
                match+=1
            elif (xx==close):
                match-=1
                if (match<=0):
                    return result
        return 0

    def gotoLine(self,event=None):
        """Ask User to go to line no"""
        lno=int(self.textbox.index(CURRENT).split(".")[0])-1
        lno=simpledialog.askinteger("MooCoderPy","Go to Line No:",initialvalue=lno)
        if lno==None:
            return
        lno=min(max(lno,0),self.lastLine())
        mark="%d.0"%(lno+1)
        self.textbox.mark_set(INSERT,mark)
        self.textbox.see(mark)
    
    def find(self,event=None):
        """Find text in editor"""
        findReplaceSettings.isreplace=False
        FindReplaceDialog(self,"Find...")
        self.textbox.focus_set()
        if findReplaceSettings.go:
            self.doFind()
    
    def findAgain(self,event=None):
        """Repeat find."""
        if findReplaceSettings.search=="":
            self.find()
        else:
            findReplaceSettings.go=True
            self.doFind()

    def doReplace(self,event:Event=None):
        findReplaceSettings.isreplace=True
        try:
            selected=self.textbox.index(SEL_FIRST)
        except:
            selected=False
        findReplaceSettings.selection=bool(selected)
        FindReplaceDialog(self,"Replace...")
        if findReplaceSettings.go:
            self.doFind()
        self.textbox.focus_set()
        return "break"

    def doFind(self,event=None):
        """Perform find command"""
        search=findReplaceSettings.search
        options={"backwards":findReplaceSettings.backward}
        options["nocase"]=not findReplaceSettings.caseSensitive
        if findReplaceSettings.wordmatch:
            options["regexp"]=True
            search=r"\y"+(re.sub(r"([\\^.?*<>\[\]$])",r"\\\1",search))+r"\y"

        replaceall=findReplaceSettings.all and findReplaceSettings.isreplace
        if replaceall:
            if findReplaceSettings.selection:
                start=SEL_FIRST
                end=SEL_LAST
            else:
                start="1.0"
                end=END
        foundonce=False
        while True:
            if replaceall:
                found=self.textbox.search(search,start,end,**options)
            elif findReplaceSettings.backward:
                ix=self.textbox.index(INSERT+" -1 chars")
                found=self.textbox.search(search, ix,"1.0",**options)
                if not(found):
                    found=self.textbox.search(search, END,ix,**options)
            else:            
                ix=self.textbox.index(INSERT)+"+1 chars" 
                found=self.textbox.search(search, ix,END,**options)
                if not(found):
                    found=self.textbox.search(search, "1.0",ix,**options)
            if found:
                foundonce=True
                lx=len(findReplaceSettings.search)
                if findReplaceSettings.isreplace:
                    self.textbox.delete(found,found+("+%d chars" % lx))
                    self.textbox.insert(found,findReplaceSettings.replace)
                    lx=len(findReplaceSettings.replace)
                self.textbox.mark_set(INSERT,found)
                self.textbox.see(found)
                self.textbox.focus_set()
                self.textbox.tag_remove("found","1.0",END)
                self.textbox.tag_add("found",found,found+("+%d chars" % lx))
                self.textbox.tag_configure("found",background="blue",foreground="white")
                self.lastfind=found
                if replaceall:
                    start="%s +%d chars" % (found,lx)
                else:
                    break
            else:
                if not(foundonce):
                    messagebox.showwarning("Search",findReplaceSettings.search+" not found.")
                break

    def testName(self)->str:
        return self.caption.lower().replace("*","").replace("=","_").replace("#","").replace(":","_")
    
    def setcolors(self):
        """Initialize colour tags"""
        for col in self.COLORLIST:
            self.textbox.tag_configure(col,foreground=col)
        self.textbox.tag_configure("bracket",foreground="white",background="red")
        self.textbox.tag_raise(SEL) # Selection should have priority.

    def lastLine(self)->int:
        """Highest line no of edit"""
        return int(self.textbox.index("end").split(".")[0])

    def getLine(self,lno:int)->str:
        """Return a selected line number."""
        return self.textbox.get(str(lno)+".0",str(lno)+".end")
    
    def cleartags(self,lno:int)->None:
        """Remove tags from a line"""
        for tag in self.COLORLIST:
            self.textbox.tag_remove(tag,str(lno)+".0",str(lno)+".end")
    
    def cleartagRegion(self,startno,endno:int):
        """Remove tags from multiple lines"""
        for tag in self.COLORLIST:
            self.textbox.tag_remove(tag,str(startno)+".0",str(endno)+".end")
    
    def highlight(self,clear:bool=False):
        """Syntax highlight all visible lines"""
        topline=int(self.textbox.index("@0,0").split(".")[0])
        bottomline=int(self.textbox.index("@0,%d" % self.textbox.winfo_height()).split(".")[0])
        if clear:
            self.cleartagRegion(topline,bottomline)
            self.syntaxlist.clear()
            self.textbox.edit_modified(False)
        for i in range(topline,bottomline+1):
            if not(i in self.syntaxlist):
                self.syntaxHighlight(i,False)
                self.syntaxlist.add(i)
        self.trackLocation()

    def syntaxHighlight(self, lno:int, clearline:bool=True):
        line=""
        x:int=0
        
        def nextWord()->str:
            nonlocal x
            result=''
            while x<len(line):
                c=line[x]
                if c in ' "()[]+-/\\*:.\{\}': 
                    if result=="":
                        result=c
                        x+=1
                    return result
                result+=c
                x+=1
            return result
        
        if (lno>=self.lastLine()) or (lno<1):
              return
        if clearline:              
            self.cleartags(lno)
        if not(self.syntax): return
        if lno==0 and self.mode==self.MODE_CODE: return # Leave programming line alone.
        line=self.getLine(lno)

        isquote=False
        isescaped=False
        isverb=False
        isproperty=True
        wordonly=False
        x=0
        color="white"
        nextcolor="white"
        while (x<len(line)-1):
            startx=x
            myword=nextWord()
            if (isescaped):
                isescaped=False
            elif isquote:
                color="yellow"
                if myword=="\\":
                    isescaped=True
                    color="orange"
                elif myword=='"':
                    isquote=False
                    wordonly=True
            elif (myword=='"'):
                isquote=True;
                color="yellow"
            else:
                if myword==':': nextcolor="magenta"
                elif myword=='.': nextcolor="lawn green"
                elif (nextcolor!="white"):
                    color=nextcolor
                    nextcolor="white"
                    wordonly=True
                elif myword.lower() in self.KEYWORDLIST: 
                    color="cyan"
                    wordonly=True
            self.setSynColor(color,lno,startx,x)
            if wordonly:
                color="white"
            wordonly=False
    
    def setSynColor(self,color:str, lno:int, start:int, end:int)->None:
        fromix="%d.%d" % (lno,start)
        toix="%d.%d" % (lno,end)
        self.textbox.tag_add(color,fromix,toix)

if __name__=="__main__":
    root=Tk()
    root.title("Test Code Editor")
    c=CodeText(root,CodeText.MODE_CODE, background="black",foreground="white",font=("Courier",12,"bold"),insertbackground="white", nowrap=True)
    c.pack(fill=BOTH,expand=True)
    c.syntax=True
#    c.setLabel("Testing label")
    with open("c:/kev/test.moo","r") as f:
        data=f.read();
        c.setText(data)
    root.mainloop()




#!/usr/bin/python3

from tkinter import *
from tkinter import filedialog
from tkinter import ttk 
from tkinter import font
from tkinter import scrolledtext
from tkinter import simpledialog
import os,sys,re,webbrowser
# Stupid packaging messes with paths...
SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
print("SCRIPT_DIR=",SCRIPT_DIR)
if not(SCRIPT_DIR in sys.path):
    sys.path.append(SCRIPT_DIR)
try:
    import MooCoderPy # This just picks up the version no.
    __VERSION__=MooCoderPy.__VERSION__
except:
    pass

from ScrollText import *
from TerminalWindow import *
import SettingsDialog


tw:TerminalWindow
lastpage:Widget
fontsize:int=12
nb:Notebook

def doOpen():
    tf = filedialog.askopenfilename(
        #initialdir="C:/td/srtmoo/", 
        title="Open Text file", 
        filetypes=(("MOO source Files", "*.moo"),("Text","*.txt"),("Any","*.*"))
        )
    if tf:
        with open(tf,"r") as f:
            data=f.read()
            tw.openEdit("Local Edit",os.path.abspath(tf),data)

def doScratchPad():
    tw.openEdit("Scratchpad","##ScratchPad##","")

def doClose():
    tw.disconnect()
    tw.saveSettings()
    root.quit()    

def doSave():
    global nb
    page=nb.nametowidget(nb.select())
    options={"title":"Save to file"}
    options["filetypes"]=(("MOO source Files", "*.moo"),("Text","*.txt"),("Any","*.*"))
    options["defaultextension"]=".moo"
    if isinstance(page,CodeText):
        if (page.mode==CodeText.MODE_EDIT and page.caption=="Local Edit"):
            options["initialfile"]=page.upload
    elif not isinstance(page,TerminalWindow):
        messagebox.showwarning("MooCoderPy","Can't save this sort of window.")
        return
    tf = filedialog.asksaveasfilename(**options)
    if not tf:
        return
    with open(tf,"w") as f:
        f.write(page.getText())
    messagebox.showinfo("MooCoderPy","Saved as "+tf)

def doConnect():
    inifile=SettingsDialog.getConfig()
    server=inifile["settings"]["Server"]
    connectstr=inifile["settings"]["Connect"]
    (server,port)=server.split(":")
    tw.connectString=connectstr
    tw.doConnect(server,int(port))
    nb.select(tw)

def doDisconnect():
    tw.disconnect()

def doSettings():
    global fontsize,nb
    oldsize=fontsize
    SettingsDialog.SettingsDialog(root,title="Server Configuration")
    getInitalSettings()
    if (fontsize!=oldsize):
        for tab in nb.tabs():
            w=nb.nametowidget(tab)
            if hasattr(w,"setFontSize"):
                w.setFontSize(fontsize)


def doupdate(event=None):
    re=tw.currentPage()
    if re and re.tabtype in (CodeText.MODE_EDIT,CodeText.MODE_CODE,CodeText.MODE_PROPERTY):
        tw.docompile(re)

def doTabChanged(event):
    global lastpage
    tabname=nb.tab("current","text")
    if (tabname=="Terminal"):
        tw.sendEntry.focus()
    #track previous page for toggle.
    w=nb.nametowidget(nb.select())
    if w!=tw:
        lastpage=w

def donewtab(event=None):
    """Load a verb into a new tab"""
    verb=simpledialog.askstring("Verb","Enter Object:Verb")
    if (verb):
        tw.loadVerb(verb)

def viewStack(event=None):
    """Make stack visible."""
    global stackloaded
    mainpack(not stackloaded)

def togglePage(event:Event=None):
    """Flip between Terminal window and previous view"""
    global nb,tw,lastpage
    w=nb.nametowidget(nb.select())
    if w==tw:
        nb.select(lastpage)
    else:
        nb.select(tw)

def doHelp(event:Event=None):
    webbrowser.open("https://github.com/rjmatthews62/MooCoderPy#readme")

def buildMenu():
    menubar=Menu(root)
    filemenu=Menu(menubar,tearoff=0)
    filemenu.add_command(label="Open", command=doOpen,underline=0)
    filemenu.add_command(label="Connect", command=doConnect,underline=0)
    filemenu.add_command(label="Disconnect", command=doDisconnect, underline=0)
    filemenu.add_command(label="Save", command=doSave, underline=0)
    filemenu.add_command(label="Exit", command=doClose,underline=1)
    settingmenu=Menu(menubar,tearoff=0)
    settingmenu.add_command(label="Server Config",command=doSettings, underline=0)
    editmenu=Menu(menubar,tearoff=0)
    editmenu.add_command(label="Send Update F5", command=doupdate, underline=0)
    editmenu.add_command(label="Scratchpad", command=doScratchPad,underline=7)
    root.bind("<F5>",doupdate)
    projectmenu=Menu(menubar,tearoff=0)
    projectmenu.add_command(label="New Tab Ctrl+N",command=donewtab,underline=0)
    projectmenu.add_command(label="Get Verbs Ctrl+Shift+V",command=tw.getVerbs,underline=0)
    projectmenu.add_command(label="Get Properties",command=tw.getProperties,underline=4)
    projectmenu.add_command(label="Get Verbs and Properties",command=tw.getBoth,underline=7)
    projectmenu.add_command(label="Clear Project",command=tw.clearProject,underline=0)
    
    root.bind("<Control-Shift-Key-V>",tw.getVerbs)
    root.bind("<Control-Shift-Key-v>",tw.getVerbs)
    # Apparently the key binding is case sensitive...
    root.bind("<Control-Key-N>",donewtab)
    root.bind("<Control-Key-n>",donewtab)
    menubar.add_cascade(label="File",menu=filemenu, underline=0)
    menubar.add_cascade(label="Settings", menu=settingmenu, underline=0)
    menubar.add_cascade(label="Edit", menu=editmenu, underline=0)
    menubar.add_cascade(label="Project",menu=projectmenu, underline=0)
    viewmenu=Menu(menubar,tearoff=0)
    viewmenu.add_command(label="Stack", command=viewStack,underline=0)
    viewmenu.add_command(label="Toggle View Ctrl+Shift+T", command=togglePage, underline=0)
    viewmenu.add_command(label="Help F1",command=doHelp)
    root.bind("<F1>",doHelp)
    root.bind("<Control-Shift-Key-T>",togglePage)
    root.bind("<Control-Shift-Key-t>",togglePage)
    menubar.add_cascade(label="View",menu=viewmenu,underline=0)
    root.config(menu=menubar)

def mainpack(usestack:bool):
    """Pack the main window so it's possible to hide and reload stack"""
    global stackloaded
    try:
        pane.forget(stack)
    except:
        pass
    if usestack:
        pane.add(stack,weight=1)
    stackloaded=usestack

def loadSettings():
    global fontsize
    ifile=SettingsDialog.getConfig()
    tw.dumpobject=ifile['settings'].get('LastDump',"")

def getInitalSettings():
    global fontsize
    ifile=SettingsDialog.getConfig()
    fontsize=ifile['settings'].getint("fontsize",12)

def doVerbDblClick(event=None):
    nd=verblist.item(verblist.selection())
    if nd["text"]:
        tw.loadVerb(nd["text"]+":"+nd["values"][1])

def doPropDblClick(event=None):
    nd=proplist.item(proplist.selection())
    if nd["text"]:
        tw.editProp(nd["text"]+"."+nd["values"][1])

def doStackClick(event:Event):
    tw.gotoError(stack)

def treeview_find(event:Event):
    findReplaceSettings.isreplace=False
    FindReplaceDialog(event.widget,"Search List")
    if findReplaceSettings.go:
        if not treeview_dofind(event.widget,False):
            treeview_findloop(event.widget)


def treeview_findloop(tv:ttk.Treeview):
    """Loop around if not found."""
    items=tv.get_children("")
    if len(items)<1: return
    if findReplaceSettings.backward:
        tv.selection_set(items[-1])
    else:
        tv.selection_set(items[0])
    treeview_dofind(tv,False)

def treeview_findagain(event:Event):
    if not findReplaceSettings.search:
        treeview_find(event)
    else:
        if not treeview_dofind(event.widget,True):
            treeview_findloop(event.widget)

def treeview_dofind(tv:ttk.Treeview,again:bool):
    items=tv.get_children("")
    ix=tv.index(tv.selection())
    if again:
        ix=ix-1 if findReplaceSettings.backward else ix+1
    cs=findReplaceSettings.caseSensitive
    search=findReplaceSettings.search
    search=search if cs else search.lower()
    if findReplaceSettings.wordmatch:
        wm=re.compile(r"\b"+search+r"\b")
    if findReplaceSettings.backward:
        items=list(items[0:ix])
        items.reverse()
    else:
        items=items[ix:]
            
    for item in items:
        v=tv.item(item)
        for vv in v["values"]:
            vv=str(vv) if cs else str(vv).lower()
            if findReplaceSettings.wordmatch:
                found=wm.search(vv)
            else:
                found=search in vv
            if found:
                tv.selection_set(item)
                tv.focus(item)
                tv.see(item)
                return True
    return False

def treeview_sort_column(tv, col, reverse,ix=-1):
    if col=="#0":
        l = [(tv.item(k, "text"), k) for k in tv.get_children('')]
        l.sort(key=lambda x:int(x[0][1:]),reverse=reverse)
    else:
        l = [(tv.set(k, col), k) for k in tv.get_children('')]
        l.sort(reverse=reverse)

    # rearrange items in sorted positions
    for index, (val, k) in enumerate(l):
        tv.move(k, '', index)

    # reverse sort next time
    reverse=not reverse
    tv.heading(col, command=lambda: \
               treeview_sort_column(tv, col, reverse))

def doBindList(tv:ttk.Treeview):
    """Setup up sort and find on a listbox."""
    for col in ["#0"]+list(tv.cget("columns")):
        tv.heading(col, command=lambda c=col:treeview_sort_column(tv, c, False,i))
    tv.bind("<Control-F>",treeview_find)
    tv.bind("<Control-f>",treeview_find)
    tv.bind("<F3>",treeview_findagain)

if not("__VERSION__" in globals()):
    __VERSION__ = "code only"

print("MooCoderPy",__VERSION__)
root = Tk()
try:
    root.iconbitmap(SCRIPT_DIR+"/moocoder.ico")
except:
    print("Couldn't find icon file. Trying png")
    try:
        ico=PhotoImage(file=SCRIPT_DIR+"/moocoder.png")
        root.iconphoto(False,ico)
    except:
        print("Didn't need that dumb icon anyway.")

myfont=font.Font(family="Arial",size=10)
getInitalSettings()
root.option_add( "*font", myfont)
root.title("MooCoderPy "+__VERSION__)
root.protocol("WM_DELETE_WINDOW",doClose)
pane=ttk.PanedWindow(root,orient=HORIZONTAL)
pane.pack(fill=BOTH, expand=True)

stack=Text(pane,width=50)
stack.bind("<Double-Button-1>",doStackClick)
nb=ttk.Notebook(pane)
pane.add(nb,weight=5)
nb.style=ttk.Style(root)
nb.style.configure("TNotebook.Tab",font=myfont)
mainpack(False)

tw=TerminalWindow(nb,background="black",foreground="white",font=("Courier",fontsize,"bold"),
                  insertbackground="white",wrap="word",takefocus=False)

tw.normalfont=myfont
tw.setFontSize(fontsize)
tw.setstackvisible=mainpack
tw.stack=stack
nb.add(tw,text="Terminal",compound=TOP,image=tw.imgList["terminal"])
verbframe=Frame(nb)
nb.add(verbframe,text="Verbs",compound=TOP,image=tw.imgList["verblist"])
lastpage=verbframe
verblist=ttk.Treeview(verbframe,columns=("c1","c2","c3","c4"))
verblist.heading("#0",text="Obj")
verblist.heading("c1",text="Name")
verblist.heading("c2",text="Verb")
verblist.heading("c3",text="Args")
verblist.heading("c4",text="Detail")
verblist.column("#0",stretch=False,width=50)
for i in range(3):
    verblist.column(i,stretch=False,width=80)
verblist.pack(fill=BOTH, expand=True)
verblist.bind("<Double-Button-1>",doVerbDblClick)
verblist.bind("<Return>",doVerbDblClick)
verblist.style=ttk.Style(verblist)
verblist.style.configure("Treeview",font=myfont)
verblist.style.configure("Treeview.Heading",font=myfont)
doBindList(verblist)

propframe=Frame(nb)
nb.add(propframe,text="Properties",compound=TOP,image=tw.imgList["proplist"])

proplist=ttk.Treeview(propframe,columns=("name","prop","detail"))
proplist.heading("#0",text="Obj")
proplist.heading("name",text="Name")
proplist.heading("prop",text="Property")
proplist.heading("detail",text="Detail")
proplist.style=verblist.style
proplist.pack(fill=BOTH, expand=True)
proplist.column("#0",stretch=False,width=50)
proplist.column(0,stretch=False,width=50)
proplist.column(1,stretch=False,width=50)
tw.fitListContents(proplist)
proplist.bind("<Double-Button-1>",doPropDblClick)
doBindList(proplist)
proplist.bind("<Return>",doPropDblClick)

nb.enable_traversal()
tw.lvVerbs=verblist
tw.lvProperties=proplist
nb.bind("<<NotebookTabChanged>>",func=doTabChanged)
nb.select(tw)
tw.pages=nb
buildMenu()
loadSettings()
try:
    root.state("zoomed") # Works in windows (and MacOs?)
except:
    root.attributes("-zoomed",True) # Works for linux

root.mainloop()
print("Done")




from configparser import ConfigParser
import os,sys
from tkinter import *
from tkinter import font,simpledialog
import SettingsDialog

class Connections(Toplevel):
    inifile:ConfigParser
    list:Listbox
    connectionList=[]
    currentConnection:StringVar
    currentLogin:StringVar
    currentIndex:int=-1

    def __init__(self,*args, **kwargs):
        y:int=0
        super().__init__(*args, **kwargs)
        self.option_add('*Font', ("Arial",10,font.BOLD))
        self.grab_set()
        f0=Frame(self)
        f0.pack(side=LEFT,fill=Y)
        lb0=Label(f0,text="Connections")
        lb0.pack(side=TOP)
        self.list=Listbox(f0,exportselection=False)
        self.list.pack(fill=Y)
        self.list.bind("<<ListboxSelect>>",self.onSelected)
        f2=Frame(f0)
        Button(f2,text="Add",command=self.addConnection).grid(column=0,row=0)
        Button(f2,text="Delete").grid(column=1,row=0)
        Button(f2,text="Rename").grid(column=2,row=0)
        f2.pack(side=BOTTOM,pady=4)
    
        f1=Frame(self)
        f1.pack(side=LEFT,fill=BOTH, expand=True)

        def addrow(caption:str, edit:Widget):
            nonlocal y
            lbl=Label(f1,text=caption)
            lbl.grid(row=y,column=0,sticky=W,padx=(8,0))
            edit.grid(row=y,column=1,sticky=W, padx=8)
            y+=1
        
        self.currentConnection=StringVar(self,value="srtmoo:8492")
        self.currentLogin=StringVar(self,value="connect robbie SRTDumpling")
        self.currentName=StringVar(self,value="Default")
        addrow("Name",Label(f1,textvariable=self.currentName))
        addrow("Address",Entry(f1,width=60,textvariable=self.currentConnection))
        addrow("Login",Entry(f1,width=60,textvariable=self.currentLogin))
        self.title("Settings")
        self.loadSettings()
    
    def addConnection(self):
        result=simpledialog.askstring("Connections","New connection name")
        if not(result):
            return
        self.addList(result,"server:port","login goes here")
        self.list.select_set(len(self.connectionList)-1)
        self.onSelected(None)

    def loadSettings(self):
        self.inifile=SettingsDialog.getConfig()
        settings=self.inifile["settings"]
        address=settings.get("Server","example.server.net:8922")
        login=settings.get("Connect","login command here")

        self.addList("Default",address,login)
        self.addList("Testing 1","testing.com:8339","connect as me")
        self.addList("SRTMoo","strmoo.net:8492","connect robbie SRTDumpling")
        self.addList("Local server","localhost:8492","connect wizard")
        self.list.select_set(0)
    
    def addList(self,name,address,login):
        self.connectionList.append((name,address,login))
        self.refreshList()
    
    def refreshList(self):
        self.list.delete(0,END)
        for x in self.connectionList:
            self.list.insert(END,x[0])

    def onSelected(self,e:Event):
        i=self.list.curselection()[0]
        if self.currentIndex>=0:
            row=self.connectionList[self.currentIndex]
            self.connectionList[self.currentIndex]=(row[0],self.currentConnection.get(),self.currentLogin.get())
        con=self.connectionList[i]
        self.currentName.set(con[0])
        self.currentConnection.set(con[1])
        self.currentLogin.set(con[2])
        self.currentIndex=i

if __name__=="__main__":
    root=Tk()
    root.title("Root")
    root.after(50,lambda: Connections(root))
    root.mainloop()


class ParseText:
    line:str
    isquote=False
    isescaped=False
    x:int=0
    c:str

    def __init__(self, line):
        self.line=line
    
    def next(self)->bool:
        if self.x>=len(self.line):
            return False
        c=self.line[self.x]
        self.x+=1
        if self.isquote:
            if self.isescaped:
                self.isescaped=False
            elif c=='"':
                self.isquote=False
            elif c=="\\":
                self.isescaped=True
        else:
            if c=='"' and not self.isescaped:
                self.isquote=True
        self.c=c
        return True


def splitList2(line):
    def indent():
        return "\n"+(" "*pad)

    p=ParseText(line)
    result=""
    pad=0
    while (p.next()):
        c=p.c
        if p.isquote:
            result+=c
        else:
            if c==",":
                result+=c+indent()
            elif c in r"[{":
                pad+=2
                result+=c+indent()
            elif c in "}]":
                pad -=2
                result+=indent()+c
            elif c==" ":
                continue
            else:
                result+=c
    return result

def joinList(data:str):
    result=""
    p=ParseText(data)
    while p.next():
        c=p.c
        if p.isquote or not(c.isspace()):
            result+=c
            if (c==","):
                result+=" "
    return result

def splitList(line):
    isquote=False
    isescaped=False
    isclosing=False
    pad=0
    result=""
    prevchar=""
    hasclosed=True
    for c in line:
        if isquote:
            if isescaped:
                isescaped=False
            elif c=='"':
                isquote=False
            elif c=="\\":
                isescaped=True
        else:
            if c=='"' and not isescaped:
                isquote=True
        if isquote:
            result+=c
        else:
            if c in r"[{":
                if not(hasclosed):
                    result+="\n"+(" " * pad)
                result+=c
                isclosing=False
                hasclosed=False
                pad+=2
            elif c in "}]":
                pad -=2
                if (isclosing):
                    result+="\n"+(" " * pad)+c
                    hasclosed=True
                else:
                    isclosing=True
                    result+=c
            elif c==",":
                if (isclosing):
                    result+=c+"\n"+(" " * pad)
                    hasclosed=True
                else:
                    isclosing=False
                    result+=c
            elif c==" " and prevchar==",":
                continue
            else:
                result+=c
                isclosing=False
            prevchar=c
    return result

if __name__=="__main__":
    s='{{["damage" -> 0.6, "name" -> "Simple", "percent" -> 75], ["damage" -> 0.7, "name" -> "Basic", "percent" -> 60], ["damage" -> 0.8, "name" -> "Standard", "percent" -> 45], ["damage" -> 1.0, "name" -> "Moderate", "percent" -> 30], ["damage" -> 1.0, "name" -> "Difficult", "percent" -> 15], ["damage" -> {1.25,2,5}, "name" -> "Direct", "percent" -> 0], ["damage" -> 1.5, "name" -> "Taxing", "percent" -> -15],{1,2,3}}}'
    print(splitList2(s))
    print(splitList('1'))
    print(splitList2('{"I am a string", "List of", "Strings"}'))
    print("Rejoined: ",joinList(splitList2(s)))



#!/usr/bin/env python

import sys,os
from distutils.core import setup
# MOST of this is handled in setup.cfg.
# Dragging version from __init__.py
fullpath=os.path.abspath("src")
if not fullpath in sys.path:
    sys.path.insert(1,fullpath)
import MooCoderPy

setup(
    version=MooCoderPy.__VERSION__
     )


from tkinter import *
from tkinter import simpledialog


class MyQueryString(simpledialog._QueryDialog):
    width=50
    def __init__(self,*args, **kw):
        if "width" in kw:
            self.width=kw["width"]
            del kw["width"]
        super().__init__(*args,**kw)

    def body(self, master):

        w = Label(master, text=self.prompt, justify=LEFT)
        w.grid(row=0, padx=5, sticky=W)

        self.entry = Entry(master, name="entry", width=self.width)
        self.entry.grid(row=1, padx=5, sticky=W+E)

        if self.initialvalue is not None:
            self.entry.insert(0, self.initialvalue)
            self.entry.select_range(0, END)

        return self.entry
    
    def getresult(self):
        return self.entry.get()

def askstring(title, prompt, **kw):
    d = MyQueryString(title, prompt, **kw)
    return d.result

if __name__=="__main__":
    root=Tk()
    s=askstring("Testing","Prompt",initialvalue="The quick brown fox jumped over the lazy dog.",width=50)
    print(s)


from tkinter import *
from tkinter import ttk
from tkinter import filedialog

def doctrl(event:Event):
    print(event)
    print(text.index("end -100 lines"))
    return "break"

def toggle():
    global showlist
    if showlist:
        pane.forget(listbox)
    else:
        pane.add(listbox)
    showlist=not(showlist)


root=Tk()
pane=ttk.PanedWindow(root,orient=HORIZONTAL)
text=Text(pane,foreground="white",background="black")
pane.add(text)
listbox=Listbox(pane)
for v in ("This","Tsting a longer line","And a really other and much longer line"):
    listbox.insert(END,v)
pane.add(listbox)
pane.pack(fill=BOTH,expand=True)
# text.pack(fill=BOTH, expand=True)
text.insert("1.0","Mary had a little lamb\nIt's fleece was white as snow.")
text.tag_configure("thing",background="steel blue", foreground="white")
print(text["selectbackground"])
text.tag_add("thing","1.0","1.end")
text.bind("<Control-Key-h>",doctrl)
text.bind("<Control-Key-H>",doctrl)
menubar=Menu(root)
editmenu=Menu(menubar)
editmenu.add_command(label="Toggle",command=toggle)
menubar.add_cascade(label="Edit",menu=editmenu)
root.config(menu=menubar)
showlist=True
root.mainloop()
